// MP3 Tag Converter
// Written by David J. Stein, Winter 2004. See "License.html" for terms and conditions of use.
//
// This module contains the front end and core functionality of the MP3 Tag Converter tool.

#region Using declarations

using System;
using System.Drawing;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.IO;
using System.Windows.Forms;
using System.Threading;

#endregion

namespace MP3TagConverterNS {

	public class MP3TagConverter : System.Windows.Forms.Form {

		#region Variables, initialization code, and Main

		private IContainer components;
		private System.Windows.Forms.TextBox txtArtist;
		private System.Windows.Forms.TextBox txtTrackName;
		private System.Windows.Forms.Label label1;
		private System.Windows.Forms.Label label2;
		private System.Windows.Forms.Label label3;
		private System.Windows.Forms.Label label4;
		private System.Windows.Forms.TextBox txtAlbum;
		private System.Windows.Forms.Button btnApply;
		private System.Windows.Forms.Label label5;
		private System.Windows.Forms.Label label6;
		private System.Windows.Forms.ListBox lstParameters;
		private System.Windows.Forms.ListBox lstAvailableParameters;
		private System.Windows.Forms.Label label7;
		private System.Windows.Forms.Label label8;
		private System.Windows.Forms.CheckBox checkOptional;
		private System.Windows.Forms.Button btnAdd;
		private System.Windows.Forms.Button btnMoveUp;
		private System.Windows.Forms.Button btnMoveDown;
		private System.Windows.Forms.Button btnDelete;
		private System.Windows.Forms.Button btnProcessSelectedFiles;
		private System.Windows.Forms.Button btnProcessFileFolder;
		private System.Windows.Forms.ListBox lstProgress;
		private System.Windows.Forms.ProgressBar pbProgress;
		private System.Windows.Forms.TextBox txtComment;
		private System.Windows.Forms.NumericUpDown nudTrackNumber;
		private System.Windows.Forms.NumericUpDown nudYear;
		private System.Windows.Forms.NumericUpDown nudTrackTotal;
		private System.Windows.Forms.Label label9;
		private System.Windows.Forms.Label label10;
		private System.Windows.Forms.NumericUpDown nudGenre;
		private System.Windows.Forms.Label label11;
		private System.Windows.Forms.ComboBox cmbGenre;
		private System.Windows.Forms.GroupBox grpProgress;
		private System.Windows.Forms.Button btnAbort;
		private System.Windows.Forms.Label label12;
		private System.Windows.Forms.TextBox txtFilename;
		private System.Windows.Forms.Label label13;
		private System.Windows.Forms.GroupBox groupBox4;
		private System.Windows.Forms.Label label14;
		private System.Windows.Forms.Label label16;
		private System.Windows.Forms.TabControl tabControl1;
		private System.Windows.Forms.TabPage tabSetTagsManually;
		private System.Windows.Forms.TabPage tabSetTagsFromFilenames;
		private System.Windows.Forms.Button btnBrowseFile;
		private System.Windows.Forms.Button btnProcessFile;
		private System.Windows.Forms.Button btnProcessFolder;
		private System.Windows.Forms.Button btnBrowseFolder;
		private System.Windows.Forms.TextBox txtFolder;
		private System.Windows.Forms.RadioButton radOmitMissingArtist;
		private System.Windows.Forms.RadioButton radUnknownMissingArtist;
		private System.Windows.Forms.RadioButton radUseBlankForMissingTracks;
		private System.Windows.Forms.RadioButton radRenumberForMissingTracks;
		private System.Windows.Forms.Panel panel1;
		private System.Windows.Forms.Panel panel2;
		private System.Windows.Forms.RadioButton radVerbose;
		private System.Windows.Forms.Label label15;
		private System.Windows.Forms.RadioButton radTerse;
		private System.Windows.Forms.Button btnHelp;

		private char[] single_filename;
		private Random r;
		private Thread processing;
		private ArrayList strFilenames;
		private string strPath;
		private bool bContinue = false;
		private System.Windows.Forms.Timer timerUpdate;
		internal List<string> astrNewProgressMessages = new List<string>();
		internal int iProgress = 0;
		private TextBox txtAlbumPrefix;
		private TextBox txtPlaylistPrefix;
		private Label label18;
		private Label label17;
		private TabPage tabSetFilenamesFromTags;
		internal string strStatus = "Progress";

		public MP3TagConverter() {
			InitializeComponent();
			single_filename = new char[300];
			r = new Random((int) System.DateTime.Now.Millisecond);
		}

		[STAThread]
		static void Main() {
			Application.Run(new MP3TagConverter());
		}

		#region Uninteresting Windows Form Designer generated code

		protected override void Dispose(bool disposing) {
			if (disposing) {
				if (components != null) {
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		private void InitializeComponent() {
			this.components = new System.ComponentModel.Container();
			System.ComponentModel.ComponentResourceManager resources = new System.ComponentModel.ComponentResourceManager(typeof(MP3TagConverter));
			this.cmbGenre = new System.Windows.Forms.ComboBox();
			this.label11 = new System.Windows.Forms.Label();
			this.nudGenre = new System.Windows.Forms.NumericUpDown();
			this.label10 = new System.Windows.Forms.Label();
			this.label9 = new System.Windows.Forms.Label();
			this.nudTrackTotal = new System.Windows.Forms.NumericUpDown();
			this.nudYear = new System.Windows.Forms.NumericUpDown();
			this.nudTrackNumber = new System.Windows.Forms.NumericUpDown();
			this.txtComment = new System.Windows.Forms.TextBox();
			this.label5 = new System.Windows.Forms.Label();
			this.label6 = new System.Windows.Forms.Label();
			this.btnApply = new System.Windows.Forms.Button();
			this.txtAlbum = new System.Windows.Forms.TextBox();
			this.label4 = new System.Windows.Forms.Label();
			this.label3 = new System.Windows.Forms.Label();
			this.label2 = new System.Windows.Forms.Label();
			this.label1 = new System.Windows.Forms.Label();
			this.txtTrackName = new System.Windows.Forms.TextBox();
			this.txtArtist = new System.Windows.Forms.TextBox();
			this.btnAbort = new System.Windows.Forms.Button();
			this.grpProgress = new System.Windows.Forms.GroupBox();
			this.btnHelp = new System.Windows.Forms.Button();
			this.label15 = new System.Windows.Forms.Label();
			this.radVerbose = new System.Windows.Forms.RadioButton();
			this.radTerse = new System.Windows.Forms.RadioButton();
			this.lstProgress = new System.Windows.Forms.ListBox();
			this.pbProgress = new System.Windows.Forms.ProgressBar();
			this.btnProcessFileFolder = new System.Windows.Forms.Button();
			this.btnProcessSelectedFiles = new System.Windows.Forms.Button();
			this.btnDelete = new System.Windows.Forms.Button();
			this.btnMoveDown = new System.Windows.Forms.Button();
			this.btnMoveUp = new System.Windows.Forms.Button();
			this.btnAdd = new System.Windows.Forms.Button();
			this.checkOptional = new System.Windows.Forms.CheckBox();
			this.label8 = new System.Windows.Forms.Label();
			this.label7 = new System.Windows.Forms.Label();
			this.lstAvailableParameters = new System.Windows.Forms.ListBox();
			this.lstParameters = new System.Windows.Forms.ListBox();
			this.label12 = new System.Windows.Forms.Label();
			this.txtFilename = new System.Windows.Forms.TextBox();
			this.btnBrowseFile = new System.Windows.Forms.Button();
			this.btnProcessFile = new System.Windows.Forms.Button();
			this.label13 = new System.Windows.Forms.Label();
			this.btnProcessFolder = new System.Windows.Forms.Button();
			this.btnBrowseFolder = new System.Windows.Forms.Button();
			this.txtFolder = new System.Windows.Forms.TextBox();
			this.groupBox4 = new System.Windows.Forms.GroupBox();
			this.txtAlbumPrefix = new System.Windows.Forms.TextBox();
			this.txtPlaylistPrefix = new System.Windows.Forms.TextBox();
			this.label18 = new System.Windows.Forms.Label();
			this.label17 = new System.Windows.Forms.Label();
			this.panel2 = new System.Windows.Forms.Panel();
			this.radUseBlankForMissingTracks = new System.Windows.Forms.RadioButton();
			this.radRenumberForMissingTracks = new System.Windows.Forms.RadioButton();
			this.label16 = new System.Windows.Forms.Label();
			this.panel1 = new System.Windows.Forms.Panel();
			this.radUnknownMissingArtist = new System.Windows.Forms.RadioButton();
			this.label14 = new System.Windows.Forms.Label();
			this.radOmitMissingArtist = new System.Windows.Forms.RadioButton();
			this.tabControl1 = new System.Windows.Forms.TabControl();
			this.tabSetTagsManually = new System.Windows.Forms.TabPage();
			this.tabSetTagsFromFilenames = new System.Windows.Forms.TabPage();
			this.tabSetFilenamesFromTags = new System.Windows.Forms.TabPage();
			this.timerUpdate = new System.Windows.Forms.Timer(this.components);
			((System.ComponentModel.ISupportInitialize) (this.nudGenre)).BeginInit();
			((System.ComponentModel.ISupportInitialize) (this.nudTrackTotal)).BeginInit();
			((System.ComponentModel.ISupportInitialize) (this.nudYear)).BeginInit();
			((System.ComponentModel.ISupportInitialize) (this.nudTrackNumber)).BeginInit();
			this.grpProgress.SuspendLayout();
			this.groupBox4.SuspendLayout();
			this.panel2.SuspendLayout();
			this.panel1.SuspendLayout();
			this.tabControl1.SuspendLayout();
			this.tabSetTagsManually.SuspendLayout();
			this.tabSetTagsFromFilenames.SuspendLayout();
			this.tabSetFilenamesFromTags.SuspendLayout();
			this.SuspendLayout();
			// 
			// cmbGenre
			// 
			this.cmbGenre.Items.AddRange(new object[] {
            "Blues (or blank)",
            "Classic Rock",
            "Country",
            "Dance",
            "Disco",
            "Funk",
            "Grunge",
            "Hip-Hop",
            "Jazz",
            "Metal",
            "New Age",
            "Oldies",
            "Other",
            "Pop",
            "R&B",
            "Rap",
            "Reggae",
            "Rock",
            "Techno",
            "Industrial",
            "Alternative",
            "Ska",
            "Death Metal",
            "Pranks",
            "Soundtrack",
            "Euro-Techno",
            "Ambient",
            "Trip-Hop",
            "Vocal",
            "Jazz+Funk",
            "Fusion",
            "Trance",
            "Classical",
            "Instrumental",
            "Acid",
            "House",
            "Game",
            "Sound Clip",
            "Gospel",
            "Noise",
            "Alternative Rock",
            "Bass",
            "Soul",
            "Punk",
            "Space",
            "Meditative",
            "Instrumental Pop",
            "Instrumental Rock",
            "Ethnic",
            "Gothic",
            "Darkwave",
            "Techno-Industrial",
            "Electronic",
            "Pop-Folk",
            "Eurodance",
            "Dream",
            "Southern Rock",
            "Comedy",
            "Cult",
            "Gangsta",
            "Top 40",
            "Christian Rap",
            "Pop/Funk",
            "Jungle",
            "Native US",
            "Cabaret",
            "New Wave",
            "Psychadelic",
            "Rave",
            "Showtunes",
            "Trailer",
            "Lo-Fi",
            "Tribal",
            "Acid Punk",
            "Acid Jazz",
            "Polka",
            "Retro",
            "Musical",
            "Rock & Roll",
            "Hard Rock",
            "Folk",
            "Folk-Rock",
            "National Folk",
            "Swing",
            "Fast Fusion",
            "Bebob",
            "Latin",
            "Revival",
            "Celtic",
            "Bluegrass",
            "Avantgarde",
            "Gothic Rock",
            "Progressive Rock",
            "Psychedelic Rock",
            "Symphonic Rock",
            "Slow Rock",
            "Big Band",
            "Chorus",
            "Easy Listening",
            "Acoustic",
            "Humour",
            "Speech",
            "Chanson",
            "Opera",
            "Chamber Music",
            "Sonata",
            "Symphony",
            "Booty Bass",
            "Primus",
            "Porn Groove",
            "Satire",
            "Slow Jam",
            "Club",
            "Tango",
            "Samba",
            "Folklore",
            "Ballad",
            "Power Ballad",
            "Rhytmic Soul",
            "Freestyle",
            "Duet",
            "Punk Rock",
            "Drum Solo",
            "Acapella",
            "Euro-House",
            "Dance Hall",
            "Goa",
            "Drum & Bass",
            "Club-House",
            "Hardcore",
            "Terror",
            "Indie",
            "BritPop",
            "Negerpunk",
            "Polsk Punk",
            "Beat",
            "Christian Gangsta Rap",
            "Heavy Metal",
            "Black Metal",
            "Crossover",
            "Contemporary Christian",
            "Christian Rock",
            "Merengue",
            "Salsa",
            "Trash Metal",
            "Anime",
            "Jpop",
            "Synthpop"});
			this.cmbGenre.Location = new System.Drawing.Point(224, 112);
			this.cmbGenre.Name = "cmbGenre";
			this.cmbGenre.Size = new System.Drawing.Size(128, 21);
			this.cmbGenre.TabIndex = 7;
			this.cmbGenre.Text = "Blues (or blank)";
			this.cmbGenre.SelectedIndexChanged += new System.EventHandler(this.cmbGenre_SelectedIndexChanged);
			// 
			// label11
			// 
			this.label11.Location = new System.Drawing.Point(112, 112);
			this.label11.Name = "label11";
			this.label11.Size = new System.Drawing.Size(96, 23);
			this.label11.TabIndex = 19;
			this.label11.Text = "or select from list:";
			this.label11.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// nudGenre
			// 
			this.nudGenre.Location = new System.Drawing.Point(56, 112);
			this.nudGenre.Maximum = new decimal(new int[] {
            255,
            0,
            0,
            0});
			this.nudGenre.Name = "nudGenre";
			this.nudGenre.Size = new System.Drawing.Size(40, 20);
			this.nudGenre.TabIndex = 3;
			this.nudGenre.ValueChanged += new System.EventHandler(this.nudGenre_ValueChanged);
			// 
			// label10
			// 
			this.label10.Location = new System.Drawing.Point(8, 112);
			this.label10.Name = "label10";
			this.label10.Size = new System.Drawing.Size(40, 23);
			this.label10.TabIndex = 17;
			this.label10.Text = "Genre";
			this.label10.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// label9
			// 
			this.label9.Location = new System.Drawing.Point(296, 48);
			this.label9.Name = "label9";
			this.label9.Size = new System.Drawing.Size(16, 23);
			this.label9.TabIndex = 15;
			this.label9.Text = "of";
			this.label9.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// nudTrackTotal
			// 
			this.nudTrackTotal.Location = new System.Drawing.Point(320, 48);
			this.nudTrackTotal.Maximum = new decimal(new int[] {
            99,
            0,
            0,
            0});
			this.nudTrackTotal.Name = "nudTrackTotal";
			this.nudTrackTotal.Size = new System.Drawing.Size(40, 20);
			this.nudTrackTotal.TabIndex = 5;
			// 
			// nudYear
			// 
			this.nudYear.Location = new System.Drawing.Point(56, 80);
			this.nudYear.Maximum = new decimal(new int[] {
            9999,
            0,
            0,
            0});
			this.nudYear.Name = "nudYear";
			this.nudYear.Size = new System.Drawing.Size(104, 20);
			this.nudYear.TabIndex = 2;
			this.nudYear.Value = new decimal(new int[] {
            2004,
            0,
            0,
            0});
			// 
			// nudTrackNumber
			// 
			this.nudTrackNumber.Location = new System.Drawing.Point(256, 48);
			this.nudTrackNumber.Maximum = new decimal(new int[] {
            99,
            0,
            0,
            0});
			this.nudTrackNumber.Name = "nudTrackNumber";
			this.nudTrackNumber.Size = new System.Drawing.Size(40, 20);
			this.nudTrackNumber.TabIndex = 12;
			this.nudTrackNumber.ValueChanged += new System.EventHandler(this.nudTrackNumber_ValueChanged);
			// 
			// txtComment
			// 
			this.txtComment.Location = new System.Drawing.Point(256, 80);
			this.txtComment.Name = "txtComment";
			this.txtComment.Size = new System.Drawing.Size(104, 20);
			this.txtComment.TabIndex = 6;
			// 
			// label5
			// 
			this.label5.Location = new System.Drawing.Point(176, 80);
			this.label5.Name = "label5";
			this.label5.Size = new System.Drawing.Size(72, 23);
			this.label5.TabIndex = 10;
			this.label5.Text = "Comment";
			this.label5.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// label6
			// 
			this.label6.Location = new System.Drawing.Point(8, 80);
			this.label6.Name = "label6";
			this.label6.Size = new System.Drawing.Size(48, 23);
			this.label6.TabIndex = 9;
			this.label6.Text = "Year";
			this.label6.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// btnApply
			// 
			this.btnApply.Location = new System.Drawing.Point(128, 152);
			this.btnApply.Name = "btnApply";
			this.btnApply.Size = new System.Drawing.Size(104, 23);
			this.btnApply.TabIndex = 8;
			this.btnApply.Text = "Apply To File...";
			this.btnApply.Click += new System.EventHandler(this.btnApply_Click);
			// 
			// txtAlbum
			// 
			this.txtAlbum.Location = new System.Drawing.Point(56, 48);
			this.txtAlbum.Name = "txtAlbum";
			this.txtAlbum.Size = new System.Drawing.Size(104, 20);
			this.txtAlbum.TabIndex = 1;
			// 
			// label4
			// 
			this.label4.Location = new System.Drawing.Point(176, 48);
			this.label4.Name = "label4";
			this.label4.Size = new System.Drawing.Size(72, 23);
			this.label4.TabIndex = 5;
			this.label4.Text = "Track No,.";
			this.label4.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// label3
			// 
			this.label3.Location = new System.Drawing.Point(8, 48);
			this.label3.Name = "label3";
			this.label3.Size = new System.Drawing.Size(48, 23);
			this.label3.TabIndex = 4;
			this.label3.Text = "Album";
			this.label3.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// label2
			// 
			this.label2.Location = new System.Drawing.Point(176, 16);
			this.label2.Name = "label2";
			this.label2.Size = new System.Drawing.Size(72, 23);
			this.label2.TabIndex = 3;
			this.label2.Text = "Track Name";
			this.label2.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// label1
			// 
			this.label1.Location = new System.Drawing.Point(8, 16);
			this.label1.Name = "label1";
			this.label1.Size = new System.Drawing.Size(48, 23);
			this.label1.TabIndex = 2;
			this.label1.Text = "Artist";
			this.label1.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// txtTrackName
			// 
			this.txtTrackName.Location = new System.Drawing.Point(256, 16);
			this.txtTrackName.Name = "txtTrackName";
			this.txtTrackName.Size = new System.Drawing.Size(104, 20);
			this.txtTrackName.TabIndex = 4;
			// 
			// txtArtist
			// 
			this.txtArtist.Location = new System.Drawing.Point(56, 16);
			this.txtArtist.Name = "txtArtist";
			this.txtArtist.Size = new System.Drawing.Size(104, 20);
			this.txtArtist.TabIndex = 0;
			// 
			// btnAbort
			// 
			this.btnAbort.Anchor = ((System.Windows.Forms.AnchorStyles) ((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Right)));
			this.btnAbort.Enabled = false;
			this.btnAbort.Location = new System.Drawing.Point(281, 191);
			this.btnAbort.Name = "btnAbort";
			this.btnAbort.Size = new System.Drawing.Size(56, 23);
			this.btnAbort.TabIndex = 3;
			this.btnAbort.Text = "Abort";
			this.btnAbort.Click += new System.EventHandler(this.btnAbort_Click);
			// 
			// grpProgress
			// 
			this.grpProgress.Anchor = ((System.Windows.Forms.AnchorStyles) ((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom)
															| System.Windows.Forms.AnchorStyles.Left)
															| System.Windows.Forms.AnchorStyles.Right)));
			this.grpProgress.Controls.Add(this.btnHelp);
			this.grpProgress.Controls.Add(this.label15);
			this.grpProgress.Controls.Add(this.radVerbose);
			this.grpProgress.Controls.Add(this.radTerse);
			this.grpProgress.Controls.Add(this.lstProgress);
			this.grpProgress.Controls.Add(this.pbProgress);
			this.grpProgress.Controls.Add(this.btnAbort);
			this.grpProgress.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.grpProgress.Location = new System.Drawing.Point(8, 232);
			this.grpProgress.Name = "grpProgress";
			this.grpProgress.Size = new System.Drawing.Size(425, 219);
			this.grpProgress.TabIndex = 14;
			this.grpProgress.TabStop = false;
			this.grpProgress.Text = "Progress";
			// 
			// btnHelp
			// 
			this.btnHelp.Anchor = ((System.Windows.Forms.AnchorStyles) ((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Right)));
			this.btnHelp.Location = new System.Drawing.Point(361, 191);
			this.btnHelp.Name = "btnHelp";
			this.btnHelp.Size = new System.Drawing.Size(56, 23);
			this.btnHelp.TabIndex = 4;
			this.btnHelp.Text = "Help";
			this.btnHelp.Click += new System.EventHandler(this.btnHelp_Click);
			// 
			// label15
			// 
			this.label15.Anchor = ((System.Windows.Forms.AnchorStyles) ((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left)));
			this.label15.Location = new System.Drawing.Point(8, 191);
			this.label15.Name = "label15";
			this.label15.Size = new System.Drawing.Size(48, 23);
			this.label15.TabIndex = 16;
			this.label15.Text = "Output:";
			this.label15.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// radVerbose
			// 
			this.radVerbose.Anchor = ((System.Windows.Forms.AnchorStyles) ((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left)));
			this.radVerbose.CheckAlign = System.Drawing.ContentAlignment.MiddleRight;
			this.radVerbose.Checked = true;
			this.radVerbose.Location = new System.Drawing.Point(136, 190);
			this.radVerbose.Name = "radVerbose";
			this.radVerbose.Size = new System.Drawing.Size(76, 24);
			this.radVerbose.TabIndex = 2;
			this.radVerbose.TabStop = true;
			this.radVerbose.Text = "Verbose";
			// 
			// radTerse
			// 
			this.radTerse.Anchor = ((System.Windows.Forms.AnchorStyles) ((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left)));
			this.radTerse.CheckAlign = System.Drawing.ContentAlignment.MiddleRight;
			this.radTerse.Location = new System.Drawing.Point(64, 190);
			this.radTerse.Name = "radTerse";
			this.radTerse.Size = new System.Drawing.Size(66, 24);
			this.radTerse.TabIndex = 1;
			this.radTerse.Text = "Terse";
			// 
			// lstProgress
			// 
			this.lstProgress.Anchor = ((System.Windows.Forms.AnchorStyles) ((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom)
															| System.Windows.Forms.AnchorStyles.Left)
															| System.Windows.Forms.AnchorStyles.Right)));
			this.lstProgress.Location = new System.Drawing.Point(8, 16);
			this.lstProgress.Name = "lstProgress";
			this.lstProgress.SelectionMode = System.Windows.Forms.SelectionMode.None;
			this.lstProgress.Size = new System.Drawing.Size(409, 147);
			this.lstProgress.TabIndex = 0;
			// 
			// pbProgress
			// 
			this.pbProgress.Anchor = ((System.Windows.Forms.AnchorStyles) (((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left)
															| System.Windows.Forms.AnchorStyles.Right)));
			this.pbProgress.Location = new System.Drawing.Point(8, 165);
			this.pbProgress.Name = "pbProgress";
			this.pbProgress.Size = new System.Drawing.Size(409, 23);
			this.pbProgress.TabIndex = 13;
			// 
			// btnProcessFileFolder
			// 
			this.btnProcessFileFolder.Anchor = System.Windows.Forms.AnchorStyles.Top;
			this.btnProcessFileFolder.Location = new System.Drawing.Point(226, 160);
			this.btnProcessFileFolder.Name = "btnProcessFileFolder";
			this.btnProcessFileFolder.Size = new System.Drawing.Size(136, 23);
			this.btnProcessFileFolder.TabIndex = 6;
			this.btnProcessFileFolder.Text = "Process Folder...";
			this.btnProcessFileFolder.Click += new System.EventHandler(this.btnProcessFileFolder_Click);
			// 
			// btnProcessSelectedFiles
			// 
			this.btnProcessSelectedFiles.Anchor = System.Windows.Forms.AnchorStyles.Top;
			this.btnProcessSelectedFiles.Location = new System.Drawing.Point(67, 160);
			this.btnProcessSelectedFiles.Name = "btnProcessSelectedFiles";
			this.btnProcessSelectedFiles.Size = new System.Drawing.Size(136, 23);
			this.btnProcessSelectedFiles.TabIndex = 5;
			this.btnProcessSelectedFiles.Text = "Process Specific Files...";
			this.btnProcessSelectedFiles.Click += new System.EventHandler(this.btnProcessSelectedFiles_Click);
			// 
			// btnDelete
			// 
			this.btnDelete.Anchor = System.Windows.Forms.AnchorStyles.Top;
			this.btnDelete.Location = new System.Drawing.Point(171, 90);
			this.btnDelete.Name = "btnDelete";
			this.btnDelete.Size = new System.Drawing.Size(72, 23);
			this.btnDelete.TabIndex = 3;
			this.btnDelete.Text = "-- Delete ->";
			this.btnDelete.Click += new System.EventHandler(this.btnDelete_Click);
			// 
			// btnMoveDown
			// 
			this.btnMoveDown.Location = new System.Drawing.Point(72, 128);
			this.btnMoveDown.Name = "btnMoveDown";
			this.btnMoveDown.Size = new System.Drawing.Size(64, 23);
			this.btnMoveDown.TabIndex = 1;
			this.btnMoveDown.Text = "Down";
			this.btnMoveDown.Click += new System.EventHandler(this.btnMoveDown_Click);
			// 
			// btnMoveUp
			// 
			this.btnMoveUp.Location = new System.Drawing.Point(8, 128);
			this.btnMoveUp.Name = "btnMoveUp";
			this.btnMoveUp.Size = new System.Drawing.Size(64, 23);
			this.btnMoveUp.TabIndex = 0;
			this.btnMoveUp.Text = "Up";
			this.btnMoveUp.Click += new System.EventHandler(this.btnMoveUp_Click);
			// 
			// btnAdd
			// 
			this.btnAdd.Anchor = System.Windows.Forms.AnchorStyles.Top;
			this.btnAdd.Location = new System.Drawing.Point(171, 58);
			this.btnAdd.Name = "btnAdd";
			this.btnAdd.Size = new System.Drawing.Size(72, 23);
			this.btnAdd.TabIndex = 2;
			this.btnAdd.Text = "<-- Add ---";
			this.btnAdd.Click += new System.EventHandler(this.btnAdd_Click);
			// 
			// checkOptional
			// 
			this.checkOptional.Anchor = ((System.Windows.Forms.AnchorStyles) ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Right)));
			this.checkOptional.CheckAlign = System.Drawing.ContentAlignment.MiddleRight;
			this.checkOptional.Location = new System.Drawing.Point(285, 128);
			this.checkOptional.Name = "checkOptional";
			this.checkOptional.Size = new System.Drawing.Size(72, 24);
			this.checkOptional.TabIndex = 4;
			this.checkOptional.Text = "Optional?";
			// 
			// label8
			// 
			this.label8.Anchor = ((System.Windows.Forms.AnchorStyles) ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Right)));
			this.label8.Location = new System.Drawing.Point(277, 8);
			this.label8.Name = "label8";
			this.label8.Size = new System.Drawing.Size(120, 23);
			this.label8.TabIndex = 3;
			this.label8.Text = "Available Parameters";
			this.label8.TextAlign = System.Drawing.ContentAlignment.MiddleCenter;
			// 
			// label7
			// 
			this.label7.Location = new System.Drawing.Point(8, 8);
			this.label7.Name = "label7";
			this.label7.Size = new System.Drawing.Size(128, 23);
			this.label7.TabIndex = 2;
			this.label7.Text = "Expected Filename";
			this.label7.TextAlign = System.Drawing.ContentAlignment.MiddleCenter;
			// 
			// lstAvailableParameters
			// 
			this.lstAvailableParameters.Anchor = ((System.Windows.Forms.AnchorStyles) ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Right)));
			this.lstAvailableParameters.Items.AddRange(new object[] {
            "Artist",
            "Track Name",
            "Album",
            "Track Number",
            "Year",
            "Genre",
            "Comment"});
			this.lstAvailableParameters.Location = new System.Drawing.Point(277, 32);
			this.lstAvailableParameters.Name = "lstAvailableParameters";
			this.lstAvailableParameters.Size = new System.Drawing.Size(120, 95);
			this.lstAvailableParameters.TabIndex = 1;
			this.lstAvailableParameters.SelectedIndexChanged += new System.EventHandler(this.lstAvailableParameters_SelectedIndexChanged);
			// 
			// lstParameters
			// 
			this.lstParameters.Items.AddRange(new object[] {
            "Artist",
            "Track Name"});
			this.lstParameters.Location = new System.Drawing.Point(8, 32);
			this.lstParameters.Name = "lstParameters";
			this.lstParameters.Size = new System.Drawing.Size(128, 95);
			this.lstParameters.TabIndex = 0;
			this.lstParameters.SelectedIndexChanged += new System.EventHandler(this.lstParameters_SelectedIndexChanged);
			// 
			// label12
			// 
			this.label12.Location = new System.Drawing.Point(8, 8);
			this.label12.Name = "label12";
			this.label12.Size = new System.Drawing.Size(80, 23);
			this.label12.TabIndex = 0;
			this.label12.Text = "Select file...";
			this.label12.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// txtFilename
			// 
			this.txtFilename.Location = new System.Drawing.Point(152, 8);
			this.txtFilename.Name = "txtFilename";
			this.txtFilename.Size = new System.Drawing.Size(202, 20);
			this.txtFilename.TabIndex = 1;
			this.txtFilename.TextChanged += new System.EventHandler(this.txtFilename_TextChanged);
			// 
			// btnBrowseFile
			// 
			this.btnBrowseFile.Location = new System.Drawing.Point(88, 8);
			this.btnBrowseFile.Name = "btnBrowseFile";
			this.btnBrowseFile.Size = new System.Drawing.Size(56, 23);
			this.btnBrowseFile.TabIndex = 0;
			this.btnBrowseFile.Text = "Browse";
			this.btnBrowseFile.Click += new System.EventHandler(this.btnBrowseFile_Click);
			// 
			// btnProcessFile
			// 
			this.btnProcessFile.Location = new System.Drawing.Point(360, 8);
			this.btnProcessFile.Name = "btnProcessFile";
			this.btnProcessFile.Size = new System.Drawing.Size(56, 23);
			this.btnProcessFile.TabIndex = 3;
			this.btnProcessFile.Text = "Process";
			this.btnProcessFile.Click += new System.EventHandler(this.btnProcessFile_Click);
			// 
			// label13
			// 
			this.label13.Location = new System.Drawing.Point(8, 40);
			this.label13.Name = "label13";
			this.label13.Size = new System.Drawing.Size(80, 23);
			this.label13.TabIndex = 4;
			this.label13.Text = "Select folder...";
			this.label13.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// btnProcessFolder
			// 
			this.btnProcessFolder.Location = new System.Drawing.Point(360, 40);
			this.btnProcessFolder.Name = "btnProcessFolder";
			this.btnProcessFolder.Size = new System.Drawing.Size(56, 23);
			this.btnProcessFolder.TabIndex = 6;
			this.btnProcessFolder.Text = "Process";
			this.btnProcessFolder.Click += new System.EventHandler(this.btnProcessFolder_Click);
			// 
			// btnBrowseFolder
			// 
			this.btnBrowseFolder.Location = new System.Drawing.Point(88, 40);
			this.btnBrowseFolder.Name = "btnBrowseFolder";
			this.btnBrowseFolder.Size = new System.Drawing.Size(56, 23);
			this.btnBrowseFolder.TabIndex = 4;
			this.btnBrowseFolder.Text = "Browse";
			this.btnBrowseFolder.Click += new System.EventHandler(this.btnBrowseFolder_Click);
			// 
			// txtFolder
			// 
			this.txtFolder.Location = new System.Drawing.Point(152, 40);
			this.txtFolder.Name = "txtFolder";
			this.txtFolder.Size = new System.Drawing.Size(202, 20);
			this.txtFolder.TabIndex = 5;
			this.txtFolder.TextChanged += new System.EventHandler(this.txtFolder_TextChanged);
			// 
			// groupBox4
			// 
			this.groupBox4.Controls.Add(this.txtAlbumPrefix);
			this.groupBox4.Controls.Add(this.txtPlaylistPrefix);
			this.groupBox4.Controls.Add(this.label18);
			this.groupBox4.Controls.Add(this.label17);
			this.groupBox4.Controls.Add(this.panel2);
			this.groupBox4.Controls.Add(this.panel1);
			this.groupBox4.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.groupBox4.Location = new System.Drawing.Point(8, 72);
			this.groupBox4.Name = "groupBox4";
			this.groupBox4.Size = new System.Drawing.Size(407, 104);
			this.groupBox4.TabIndex = 8;
			this.groupBox4.TabStop = false;
			this.groupBox4.Text = "M3U Options";
			// 
			// txtAlbumPrefix
			// 
			this.txtAlbumPrefix.Location = new System.Drawing.Point(251, 73);
			this.txtAlbumPrefix.Name = "txtAlbumPrefix";
			this.txtAlbumPrefix.Size = new System.Drawing.Size(150, 20);
			this.txtAlbumPrefix.TabIndex = 13;
			this.txtAlbumPrefix.Text = "Album - ";
			// 
			// txtPlaylistPrefix
			// 
			this.txtPlaylistPrefix.Location = new System.Drawing.Point(251, 34);
			this.txtPlaylistPrefix.Name = "txtPlaylistPrefix";
			this.txtPlaylistPrefix.Size = new System.Drawing.Size(150, 20);
			this.txtPlaylistPrefix.TabIndex = 12;
			this.txtPlaylistPrefix.Text = "Playlist - ";
			// 
			// label18
			// 
			this.label18.AutoSize = true;
			this.label18.Location = new System.Drawing.Point(248, 59);
			this.label18.Name = "label18";
			this.label18.Size = new System.Drawing.Size(68, 13);
			this.label18.TabIndex = 11;
			this.label18.Text = "Album Prefix:";
			// 
			// label17
			// 
			this.label17.AutoSize = true;
			this.label17.Location = new System.Drawing.Point(248, 20);
			this.label17.Name = "label17";
			this.label17.Size = new System.Drawing.Size(71, 13);
			this.label17.TabIndex = 10;
			this.label17.Text = "Playlist Prefix:";
			// 
			// panel2
			// 
			this.panel2.Controls.Add(this.radUseBlankForMissingTracks);
			this.panel2.Controls.Add(this.radRenumberForMissingTracks);
			this.panel2.Controls.Add(this.label16);
			this.panel2.Location = new System.Drawing.Point(112, 16);
			this.panel2.Name = "panel2";
			this.panel2.Size = new System.Drawing.Size(112, 80);
			this.panel2.TabIndex = 9;
			// 
			// radUseBlankForMissingTracks
			// 
			this.radUseBlankForMissingTracks.Location = new System.Drawing.Point(0, 48);
			this.radUseBlankForMissingTracks.Name = "radUseBlankForMissingTracks";
			this.radUseBlankForMissingTracks.Size = new System.Drawing.Size(112, 24);
			this.radUseBlankForMissingTracks.TabIndex = 5;
			this.radUseBlankForMissingTracks.Text = "Include blank";
			// 
			// radRenumberForMissingTracks
			// 
			this.radRenumberForMissingTracks.Checked = true;
			this.radRenumberForMissingTracks.Location = new System.Drawing.Point(0, 24);
			this.radRenumberForMissingTracks.Name = "radRenumberForMissingTracks";
			this.radRenumberForMissingTracks.Size = new System.Drawing.Size(112, 24);
			this.radRenumberForMissingTracks.TabIndex = 4;
			this.radRenumberForMissingTracks.TabStop = true;
			this.radRenumberForMissingTracks.Text = "Renumber tracks";
			// 
			// label16
			// 
			this.label16.Location = new System.Drawing.Point(0, 8);
			this.label16.Name = "label16";
			this.label16.Size = new System.Drawing.Size(112, 23);
			this.label16.TabIndex = 3;
			this.label16.Text = "Missing Album Track:";
			// 
			// panel1
			// 
			this.panel1.Controls.Add(this.radUnknownMissingArtist);
			this.panel1.Controls.Add(this.label14);
			this.panel1.Controls.Add(this.radOmitMissingArtist);
			this.panel1.Location = new System.Drawing.Point(8, 16);
			this.panel1.Name = "panel1";
			this.panel1.Size = new System.Drawing.Size(112, 80);
			this.panel1.TabIndex = 8;
			// 
			// radUnknownMissingArtist
			// 
			this.radUnknownMissingArtist.Checked = true;
			this.radUnknownMissingArtist.Location = new System.Drawing.Point(8, 24);
			this.radUnknownMissingArtist.Name = "radUnknownMissingArtist";
			this.radUnknownMissingArtist.Size = new System.Drawing.Size(96, 24);
			this.radUnknownMissingArtist.TabIndex = 1;
			this.radUnknownMissingArtist.TabStop = true;
			this.radUnknownMissingArtist.Text = "\"Anonymous\"";
			// 
			// label14
			// 
			this.label14.Location = new System.Drawing.Point(8, 8);
			this.label14.Name = "label14";
			this.label14.Size = new System.Drawing.Size(96, 23);
			this.label14.TabIndex = 0;
			this.label14.Text = "Missing Artist:";
			// 
			// radOmitMissingArtist
			// 
			this.radOmitMissingArtist.Location = new System.Drawing.Point(8, 48);
			this.radOmitMissingArtist.Name = "radOmitMissingArtist";
			this.radOmitMissingArtist.Size = new System.Drawing.Size(96, 24);
			this.radOmitMissingArtist.TabIndex = 2;
			this.radOmitMissingArtist.Text = "Omit artist";
			// 
			// tabControl1
			// 
			this.tabControl1.Anchor = ((System.Windows.Forms.AnchorStyles) (((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left)
															| System.Windows.Forms.AnchorStyles.Right)));
			this.tabControl1.Controls.Add(this.tabSetTagsFromFilenames);
			this.tabControl1.Controls.Add(this.tabSetTagsManually);
			this.tabControl1.Controls.Add(this.tabSetFilenamesFromTags);
			this.tabControl1.HotTrack = true;
			this.tabControl1.Location = new System.Drawing.Point(8, 8);
			this.tabControl1.Name = "tabControl1";
			this.tabControl1.SelectedIndex = 0;
			this.tabControl1.Size = new System.Drawing.Size(425, 218);
			this.tabControl1.SizeMode = System.Windows.Forms.TabSizeMode.FillToRight;
			this.tabControl1.TabIndex = 16;
			// 
			// tabSetTagsManually
			// 
			this.tabSetTagsManually.Controls.Add(this.txtTrackName);
			this.tabSetTagsManually.Controls.Add(this.label1);
			this.tabSetTagsManually.Controls.Add(this.label2);
			this.tabSetTagsManually.Controls.Add(this.label3);
			this.tabSetTagsManually.Controls.Add(this.label4);
			this.tabSetTagsManually.Controls.Add(this.txtAlbum);
			this.tabSetTagsManually.Controls.Add(this.btnApply);
			this.tabSetTagsManually.Controls.Add(this.label6);
			this.tabSetTagsManually.Controls.Add(this.label5);
			this.tabSetTagsManually.Controls.Add(this.txtComment);
			this.tabSetTagsManually.Controls.Add(this.nudTrackNumber);
			this.tabSetTagsManually.Controls.Add(this.nudYear);
			this.tabSetTagsManually.Controls.Add(this.nudTrackTotal);
			this.tabSetTagsManually.Controls.Add(this.label9);
			this.tabSetTagsManually.Controls.Add(this.label10);
			this.tabSetTagsManually.Controls.Add(this.nudGenre);
			this.tabSetTagsManually.Controls.Add(this.label11);
			this.tabSetTagsManually.Controls.Add(this.cmbGenre);
			this.tabSetTagsManually.Controls.Add(this.txtArtist);
			this.tabSetTagsManually.Location = new System.Drawing.Point(4, 24);
			this.tabSetTagsManually.Name = "tabSetTagsManually";
			this.tabSetTagsManually.Padding = new System.Windows.Forms.Padding(4);
			this.tabSetTagsManually.Size = new System.Drawing.Size(417, 190);
			this.tabSetTagsManually.TabIndex = 0;
			this.tabSetTagsManually.Text = "Manually Set Tags ";
			this.tabSetTagsManually.UseVisualStyleBackColor = true;
			// 
			// tabSetTagsFromFilenames
			// 
			this.tabSetTagsFromFilenames.Controls.Add(this.btnProcessFileFolder);
			this.tabSetTagsFromFilenames.Controls.Add(this.btnProcessSelectedFiles);
			this.tabSetTagsFromFilenames.Controls.Add(this.btnDelete);
			this.tabSetTagsFromFilenames.Controls.Add(this.btnMoveDown);
			this.tabSetTagsFromFilenames.Controls.Add(this.btnMoveUp);
			this.tabSetTagsFromFilenames.Controls.Add(this.btnAdd);
			this.tabSetTagsFromFilenames.Controls.Add(this.checkOptional);
			this.tabSetTagsFromFilenames.Controls.Add(this.label8);
			this.tabSetTagsFromFilenames.Controls.Add(this.label7);
			this.tabSetTagsFromFilenames.Controls.Add(this.lstAvailableParameters);
			this.tabSetTagsFromFilenames.Controls.Add(this.lstParameters);
			this.tabSetTagsFromFilenames.Location = new System.Drawing.Point(4, 24);
			this.tabSetTagsFromFilenames.Name = "tabSetTagsFromFilenames";
			this.tabSetTagsFromFilenames.Size = new System.Drawing.Size(417, 190);
			this.tabSetTagsFromFilenames.TabIndex = 1;
			this.tabSetTagsFromFilenames.Text = "Set Tags From Filenames ";
			this.tabSetTagsFromFilenames.UseVisualStyleBackColor = true;
			// 
			// tabSetFilenamesFromTags
			// 
			this.tabSetFilenamesFromTags.Controls.Add(this.label12);
			this.tabSetFilenamesFromTags.Controls.Add(this.label13);
			this.tabSetFilenamesFromTags.Controls.Add(this.txtFilename);
			this.tabSetFilenamesFromTags.Controls.Add(this.txtFolder);
			this.tabSetFilenamesFromTags.Controls.Add(this.btnBrowseFolder);
			this.tabSetFilenamesFromTags.Controls.Add(this.btnProcessFolder);
			this.tabSetFilenamesFromTags.Controls.Add(this.btnProcessFile);
			this.tabSetFilenamesFromTags.Controls.Add(this.groupBox4);
			this.tabSetFilenamesFromTags.Controls.Add(this.btnBrowseFile);
			this.tabSetFilenamesFromTags.Location = new System.Drawing.Point(4, 24);
			this.tabSetFilenamesFromTags.Name = "tabSetFilenamesFromTags";
			this.tabSetFilenamesFromTags.Size = new System.Drawing.Size(417, 190);
			this.tabSetFilenamesFromTags.TabIndex = 2;
			this.tabSetFilenamesFromTags.Text = "Set Filenames From Tags ";
			this.tabSetFilenamesFromTags.UseVisualStyleBackColor = true;
			// 
			// timerUpdate
			// 
			this.timerUpdate.Enabled = true;
			this.timerUpdate.Tick += new System.EventHandler(this.timerUpdate_Tick);
			// 
			// MP3TagConverter
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(441, 458);
			this.Controls.Add(this.tabControl1);
			this.Controls.Add(this.grpProgress);
			this.Icon = ((System.Drawing.Icon) (resources.GetObject("$this.Icon")));
			this.MinimumSize = new System.Drawing.Size(400, 472);
			this.Name = "MP3TagConverter";
			this.Text = "MP3 Tag Converter";
			this.Closing += new System.ComponentModel.CancelEventHandler(this.MP3ID3Converter_Closing);
			((System.ComponentModel.ISupportInitialize) (this.nudGenre)).EndInit();
			((System.ComponentModel.ISupportInitialize) (this.nudTrackTotal)).EndInit();
			((System.ComponentModel.ISupportInitialize) (this.nudYear)).EndInit();
			((System.ComponentModel.ISupportInitialize) (this.nudTrackNumber)).EndInit();
			this.grpProgress.ResumeLayout(false);
			this.groupBox4.ResumeLayout(false);
			this.groupBox4.PerformLayout();
			this.panel2.ResumeLayout(false);
			this.panel1.ResumeLayout(false);
			this.tabControl1.ResumeLayout(false);
			this.tabSetTagsManually.ResumeLayout(false);
			this.tabSetTagsManually.PerformLayout();
			this.tabSetTagsFromFilenames.ResumeLayout(false);
			this.tabSetFilenamesFromTags.ResumeLayout(false);
			this.tabSetFilenamesFromTags.PerformLayout();
			this.ResumeLayout(false);

		}
		#endregion

		#endregion

		#region Core functions

		public bool ProcessMP3(string strFilename, string strArtist, string strTrackName, string strAlbum, int iTrackNumber, int iTrackTotal, int iYear, int iGenre, string strComment) {
			// This file writes MP3 tag data.
			if (File.Exists(strFilename) == false)  // can't find source file - exit.
				return false;

			// read contents of old file
			byte[] buffer = new byte[(new FileInfo(strFilename)).Length];
			BinaryReader b = new BinaryReader((Stream) new FileStream(strFilename, FileMode.Open));
			for (long l = 0; l < buffer.Length; l++)
				buffer[l] = b.ReadByte();
			b.Close();

			// scan for lyrics info and trim out if found
			long lBegin = -1;
			long lEnd = -1;
			for (long i = 0; i < buffer.Length - 8;) {
				if (buffer[i] != 'L')
					i++;
				else if (buffer[i+1] != 'Y')
					i += 2;
				else if (buffer[i+2] != 'R')
					i += 3;
				else if (buffer[i+3] != 'I')
					i += 4;
				else if (buffer[i+4] != 'C')
					i += 5;
				else if (buffer[i+5] != 'S')
					i += 6;
				else if (buffer[i+6] != 'B') {
					if (buffer[i+6] != '2')
						i += 7;
					else if (buffer[i+7] != '0')
						i += 8;
					else if (buffer[i+8] != '0')
						i += 9;
					else {
						lEnd = i + 9;
						i += 9;
					}
				}
				else if (buffer[i+7] != 'E')
					i += 8;
				else if (buffer[i+8] != 'G')
					i += 9;
				else if (buffer[i+9] != 'I')
					i += 10;
				else if (buffer[i+10] != 'N')
					i += 11;
				else {
					lBegin = i;
					i += 1;
				}
			}
			if ((lBegin > -1) && (lEnd > lBegin)) {  // found appropriate information - create new file
				// create new buffer containing contents of old buffer with lyrics stuff removed
				byte[] newbuffer = new byte[buffer.Length - (lEnd - lBegin)];
				for (long i = 0; i < lBegin; i++)
					newbuffer[i] = buffer[i];
				for (long i = lEnd; i < buffer.Length; i++)
					newbuffer[i - (lEnd - lBegin)] = buffer[i];
				buffer = newbuffer;
			}

			// ID3v1 tag
			if ((buffer.Length < 128) || ((buffer[buffer.Length - 128] != 'T') || (buffer[buffer.Length - 127] != 'A') || (buffer[buffer.Length - 126] != 'G'))) {  // no ID3v1 tag - expand buffer and insert space for ID3v1 tag
				byte[] newbuffer = new byte[buffer.Length + 128];
				for (long i = 0; i < buffer.Length; i++)
					newbuffer[i] = buffer[i];
				newbuffer[buffer.Length - 128] = (byte) 'T';
				newbuffer[buffer.Length - 127] = (byte) 'A';
				newbuffer[buffer.Length - 126] = (byte) 'G';
				buffer = newbuffer;
			}
			else {  // wipe old ID3v1 info
				for (long i = buffer.Length - 125; i < buffer.Length; i++)
					buffer[i] = (byte) 0;
			}

			// write ID3v1 tag
			for (int i = 0; i < 30 && i < strTrackName.Length; i++)
				buffer[buffer.Length - 125 + i] = (byte) strTrackName[i];
			for (int i = 0; i < 30 && i < strArtist.Length; i++)
				buffer[buffer.Length - 95 + i] = (byte) strArtist[i];
			for (int i = 0; i < 30 && i < strAlbum.Length; i++)
				buffer[buffer.Length - 65 + i] = (byte) strAlbum[i];
			if (iYear > 0) {
				buffer[buffer.Length - 35] = (byte) (((iYear / 1000) % 10) + '0');
				buffer[buffer.Length - 34] = (byte) (((iYear / 100) % 10) + '0');
				buffer[buffer.Length - 33] = (byte) (((iYear / 10) % 10) + '0');
				buffer[buffer.Length - 32] = (byte) (((iYear / 1) % 10) + '0');
			} else {
				for (int i = 0; i < 4; i++)
					buffer[buffer.Length - 35 + i] = 0;
			}
			for (int i = 0; i < 29 && i < strComment.Length; i++)
				buffer[buffer.Length - 31 + i] = (byte) strComment[i];
			buffer[buffer.Length - 2] = (byte) (iTrackNumber % 256);
			buffer[buffer.Length - 1] = (byte) (iGenre % 256);

			// write new file with ID3V2 data
			File.Delete(strFilename);
			BinaryWriter bNewFile = new BinaryWriter(new FileStream(strFilename, FileMode.CreateNew));
			WriteStringToFile(bNewFile, "ID3");
			// write version information
			bNewFile.Write((char) 3);
			bNewFile.Write((char) 0);
			bNewFile.Write((char) 0);
			// compute and write ID3v2 tag size
			long lTagLength = 11 + strTrackName.Length;
			if (strArtist.Length > 0)
				lTagLength += 11 + strArtist.Length;
			if (strAlbum.Length > 0) {
				lTagLength += 11 + strAlbum.Length;
				if (iTrackNumber > 0)
					lTagLength += 11 + (iTrackNumber > 9 ? 2 : 1) + 1 + (iTrackTotal > 9 ? 2 : 1);
			}
			if (iGenre > 0)
				lTagLength += 14 + (iGenre > 9 ? 1 : 0) + (iGenre > 99 ? 1 : 0);
			for (int power = 3; power >= 0; power--)  // write total tag length (four bytes, most significant byte first)
				bNewFile.Write((char) ((lTagLength / Math.Pow(128, power)) % 128));   // note: ID3V2 tags don't use most significant bit, so it's only 7 bits here, not 8.
			// write tag parts
			WriteID3v2TrackPart(bNewFile, "TPE1", strArtist);
			WriteID3v2TrackPart(bNewFile, "TIT2", strTrackName);
			if (strAlbum.Length > 0) {
				WriteID3v2TrackPart(bNewFile, "TALB", strAlbum);
				if (iTrackNumber > 0)
					WriteID3v2TrackPart(bNewFile, "TRCK", iTrackNumber + "/" + iTrackTotal);
			}
			if (iGenre > 0)
				WriteID3v2TrackPart(bNewFile, "TCON", "(" + (iGenre > 99 ? "" + (char) (((iGenre / 100) % 10) + '0') : "") + (iGenre > 9 ? "" + (char) (((iGenre / 10) % 10) + '0') : "") + (char) ((iGenre % 10) + '0') + ")");
			// done with ID3v2 tag - write contents of buffer and close file.
			long start = 0;
			if ((buffer.Length >= 7) && (buffer[0] == 'I') && (buffer[1] == 'D') && (buffer[2] == '3')) {  // offset start from end of ID3v2 tag
				start = 10;
				for (int power = 3; power >= 0; power--)
					start += (long) Math.Pow(128, power) * buffer[9 - power];
			}
			for (long position = start; position < buffer.Length; position++)
				bNewFile.Write(buffer[position]);
			bNewFile.Close();
			return true;
		}

		public void ProcessTags() {
			// This function automatically processes all MP3s (writes ID3 tags) based on filename parsing options.
			bContinue = true;
			// assemble albums
			ArrayList aAlbums = AssembleAlbums(((string) strFilenames[0]).Substring(0, ((string) strFilenames[0]).LastIndexOf(@"\")));
			// prepare parameter list
			ArrayList aParameters = new ArrayList();
			foreach (string strParameter in lstParameters.Items)
				aParameters.Add(strParameter);
			// parse filenames
			int iCount = 0;
			for (; (iCount < strFilenames.Count) && (bContinue == true); iCount++) {
				string strFilename = (string) strFilenames[iCount];
				// find filename tokens
				string strFilenameTrimmed = strFilename.Substring(strFilename.LastIndexOf(@"\") + 1);
				// find album
				Album album = FindAlbum(aAlbums, strFilenameTrimmed);
				int iTrack = (album == null ? 0 : FindTrackNumber(album, strFilenameTrimmed));
				if (strFilenameTrimmed.EndsWith(".mp3"))
					strFilenameTrimmed = strFilenameTrimmed.Substring(0, strFilenameTrimmed.Length - 4);
				ArrayList aFilenameTokens = new ArrayList();
				for (int i = 0; i < strFilenameTrimmed.Length;) {
					int iIndex = strFilenameTrimmed.IndexOf(" - ", i);
					if (iIndex != -1) {
						string strSubstring = strFilenameTrimmed.Substring(i, iIndex - i).Trim();
						if (strSubstring.Length > 0)
							aFilenameTokens.Add(strSubstring);
						i = iIndex + 3;
					}
					else {
						string strSubstring = strFilenameTrimmed.Substring(i, strFilenameTrimmed.Length - i).Trim();
						if (strSubstring.Length > 0)
							aFilenameTokens.Add(strSubstring);
						i = strFilenameTrimmed.Length;
					}
				}
				if (aFilenameTokens.Count > 0) {
					ArrayList aSolutions = new ArrayList();
					RecursiveFilenameParse(aFilenameTokens, 0, aParameters, 0, "", ref aSolutions);
					if (aSolutions.Count == 1) {  // only one matching solution
						ArrayList aFinalSolution = ParseSolution(aFilenameTokens, (string) aSolutions[0]);
						this.ProcessMP3(strFilename, (string) aFinalSolution[0], (string) aFinalSolution[1], album == null ? (string) aFinalSolution[2] : album.strName, album == null ? (int) aFinalSolution[3] : iTrack, album == null ? 99 : album.iTracks, (int) aFinalSolution[4], (int) aFinalSolution[5], (string) aFinalSolution[6]);
						if (radVerbose.Checked)
							astrNewProgressMessages.Add("Tagging " + strFilenameTrimmed);
						iProgress = 100 * iCount / strFilenames.Count;
						strStatus = "Progress: " + iCount + "/" + strFilenames.Count;
					}
					else if (aSolutions.Count > 1) {   // let use choose a solution
						ChooseFilenameParseSolution c = new ChooseFilenameParseSolution(aFilenameTokens, aSolutions);
						if (c.ShowDialog() == DialogResult.OK) {
							int iSelected = c.iSelectedSolution;
							if (iSelected > -1) {  // selected a solution
								ArrayList aFinalSolution = ParseSolution(aFilenameTokens, (string) aSolutions[c.iSelectedSolution]);
								// parse out album information
								if (album != null) {
									aFinalSolution[2] = album.strName;
									aFinalSolution[3] = iTrack;
									aFinalSolution[4] = album.aTracks.Count;
								}
								this.ProcessMP3(strFilename, (string) aFinalSolution[0], (string) aFinalSolution[1], (string) aFinalSolution[2], (int) aFinalSolution[3], 99, (int) aFinalSolution[4], (int) aFinalSolution[5], (string) aFinalSolution[6]);
								if (radVerbose.Checked)
									astrNewProgressMessages.Add("Tagging " + strFilenameTrimmed);
								iProgress = 100 * iCount / strFilenames.Count;
								strStatus = "Progress: " + iCount + "/" + strFilenames.Count;
							}
						}
					}
				}
			}

			// show completion
			iProgress = 100;
			astrNewProgressMessages.Add("Finished. Total files processed: " + iCount);
			strStatus = "Progress: " + iCount + "/" + strFilenames.Count;
			// terminate thread
			Thread tempThread = processing;
			processing = null;
			tempThread.Abort();
		}

		private void ProcessFilenames() {
			// This function automatically renames all MP3s (and updates/writes M3Us) based on ID3 tag data.

			astrNewProgressMessages.Add("Reading playlists");

			ArrayList aPlaylists = ReadPlaylists(strPath);
			ArrayList aAlbums = new ArrayList();
			int count = 0;
			foreach (string filename in strFilenames) {
				Track track = ExtractInfoFromTag(filename);
				if (radVerbose.Checked)
					astrNewProgressMessages.Add("Renaming " + filename + " to " + track.strFilename);
				UpdatePlaylists(filename, track.strFilename, aPlaylists);
				if (track.strAlbum.Length != 0)
					UpdateAlbums(track, aAlbums);
				iProgress = 100 * (++count) / strFilenames.Count;
				string strNewFilename = strPath + @"\" + track.strFilename;
				if (strNewFilename != filename) {
					if (File.Exists(strNewFilename))
						File.Delete(strNewFilename);
					File.Move(filename, strPath + @"\" + track.strFilename);
				}
			}
			astrNewProgressMessages.Add("Done renaming files. Updating playlists and albums.");
			WritePlaylists(aPlaylists);
			WriteAlbums(aAlbums);
			astrNewProgressMessages.Add("Done.");
			Thread tempThread = processing;
			processing = null;
			tempThread.Abort();
		}

		private ArrayList ReadPlaylists(string path) {
			// This function reads all playlists into a list.
			ArrayList aPlaylists = new ArrayList();
			string[] M3Us = Directory.GetFiles(path, "*.m3u");
			foreach (string filename in M3Us) {
				if ((filename.IndexOf("Playlist - ") != -1) || ((txtPlaylistPrefix.Text.Length > 0) && (filename.IndexOf(txtPlaylistPrefix.Text) != -1)) || (((txtAlbumPrefix.Text.Length == 0) || (filename.IndexOf(txtAlbumPrefix.Text) == -1)) && (filename.IndexOf("Album - ") == -1) && (filename.IndexOf("Audiobook - ") == -1))) {
					Playlist playlist = new Playlist();
					playlist.strPlaylistName = filename;
					StreamReader file = new StreamReader(filename);
					string[] lines = file.ReadToEnd().Split(new char[] { '\n' });
					file.Close();
					foreach (string line in lines) {
						if ((line.Length > 3) && (line.StartsWith("#") == false) && (line.StartsWith("Missing Track.mp3") == false))
							playlist.aTracks.Add(line.Trim());
					}
					aPlaylists.Add(playlist);
				}
			}
			return aPlaylists;
		}

		private void UpdatePlaylists(string strOldFilename, string strNewFilename, ArrayList aPlaylists) {
			// This function updates all playlists after a file is renamed.
			foreach (Playlist playlist in aPlaylists) {
				for (int i = 0; i < playlist.aTracks.Count; i++) {
					string strTrack = (string) playlist.aTracks[i];
					if (strOldFilename.EndsWith(strTrack)) {
						if (radVerbose.Checked)
							astrNewProgressMessages.Add("Updating playlist " + playlist.strPlaylistName);
						playlist.aTracks.RemoveAt(i);
						playlist.aTracks.Insert(i, strNewFilename);
					}
				}
			}
		}

		private void UpdateAlbums(Track track, ArrayList aAlbums) {
			// This function updates all album playlists based on M3U data.
			bool foundAlbum = false;
			if ((track.strAlbum.Length > 0) && (track.strAlbum != "<Unknown>")) {
				string strName = track.strAlbum;
				foreach (Album album in aAlbums) {
					if (album.strName == strName) {
						if (radVerbose.Checked)
							astrNewProgressMessages.Add("Adding track to album " + strName);
						foundAlbum = true;
						if (album.iTotalTracks != track.iTrackTotal) {
							astrNewProgressMessages.Add("Warning: Album " + track.strAlbum + " has conflicting track count in track " + track.strFilename + ".");
							album.iTotalTracks = Math.Max(album.iTotalTracks, track.iTrackTotal);
						}
						int i = 0;
						for (; i < album.aTracks.Count; i++) {
							if (((Track) album.aTracks[i]).iTrackNumber > track.iTrackNumber) {  // insert here
								album.aTracks.Insert(i, track);
								album.iTracks++;
								break;
							}
							else if (((Track) album.aTracks[i]).iTrackNumber ==  track.iTrackNumber) {  // conflict
								astrNewProgressMessages.Add("Warning: Album " + track.strAlbum + " has two tracks with track #" + track.iTrackNumber + ": " + ((Track) album.aTracks[i]).strFilename + " and " + track.strFilename + ". Omitting the latter.");
								break;
							}
						}
						if (i == album.aTracks.Count)  {  // add to end of album
							album.aTracks.Add(track);
							album.iTracks++;
						}
						break;
					}
				}
				if (foundAlbum == false) {  // have to create new album
					astrNewProgressMessages.Add("Creating album " + strName);
					Album Album = new Album();
					Album.strName = strName;
					Album.iTotalTracks = track.iTrackTotal;
					Album.iTracks = 1;
					Album.aTracks.Add(track);
					aAlbums.Add(Album);
				}
			}
		}

		private void WritePlaylists(ArrayList aPlaylists) {
			// This function writes all non-album playlists.-
			foreach (Playlist playlist in aPlaylists) {
				if (radVerbose.Checked)
					astrNewProgressMessages.Add("Updating playlist " + playlist.strPlaylistName);
				StreamWriter file = new StreamWriter(playlist.strPlaylistName, false, System.Text.Encoding.ASCII);
				foreach (string strTrack in playlist.aTracks)
					file.WriteLine(strTrack);
				file.Close();
			}
		}

		private ArrayList AssembleAlbums(string path) {
			// This function assembles a list of all album M3Us in the current file folder.
			ArrayList aAlbums = new ArrayList();
			string[] strM3uFiles = Directory.GetFiles(path, "*.m3u");
			foreach (string strM3uFile in strM3uFiles) {
				if (strM3uFile.IndexOf("Playlist - ") == -1) {  // this is an album
					StreamReader m3uFile = new StreamReader(strM3uFile);
					if (m3uFile != null) {
						Album album = new Album();
						// this part is kinda kludged... but no matter
						int iHashPosition = strM3uFile.LastIndexOf(@"\");
						string strTrimmedFilename = strM3uFile.Substring(iHashPosition + 1, strM3uFile.Length - iHashPosition - 5).Trim();
						if (strTrimmedFilename.StartsWith("Album - "))
							strTrimmedFilename = strTrimmedFilename.Substring(8);
						if (strTrimmedFilename.StartsWith("Audiobook - "))
							strTrimmedFilename = strTrimmedFilename.Substring(12);
						album.strName = strTrimmedFilename;
						string strInput = m3uFile.ReadLine();
						while (strInput != null) {
							strInput = strInput.Trim();
							if ((strInput.StartsWith("#") == false) && (strInput.Length > 0))  // add track
								album.aTracks.Add(strInput);
							strInput = m3uFile.ReadLine();
						}
						m3uFile.Close();
						album.iTracks = album.aTracks.Count;
						aAlbums.Add(album);
					}
				}
			}
			return aAlbums;
		}

		private void WriteAlbums(ArrayList aAlbums) {
			// This function writes all album-tracklist playlists.
			foreach (Album aAlbum in aAlbums) {
				if (radVerbose.Checked)
					astrNewProgressMessages.Add("Updating album " + aAlbum.strName);
				StreamWriter file = new StreamWriter(strPath + @"\" + txtAlbumPrefix.Text + aAlbum.strName + ".m3u", false, System.Text.Encoding.ASCII);
				int iTrackCount = 1;
				foreach (Track track in aAlbum.aTracks) {
					if (radRenumberForMissingTracks.Checked == true) {
						while (track.iTrackNumber > iTrackCount) {
							file.WriteLine("Missing Track.mp3");
							iTrackCount++;
						}
					}
					file.WriteLine(track.strFilename);
					iTrackCount++;
				}
				if (radRenumberForMissingTracks.Checked == true) {
					while (iTrackCount < aAlbum.iTotalTracks) {
						file.WriteLine("Missing Track.mp3");
						iTrackCount++;
					}
				}
				file.Close();
			}
		}

		private Track ExtractInfoFromTag(string strFilename) {
			// This function extracts filename info from an ID3 tag.
			Track track = new Track();
			track.strArtist = (radUnknownMissingArtist.Checked ? "Anonymous" : "");
			if (File.Exists(strFilename)) {
				StreamReader file = new StreamReader(strFilename, System.Text.Encoding.ASCII);
				try {
					char[] tag = new char[10];
					file.Read(tag, 0, 10);
					if ((tag[0] == 'I') && (tag[1] == 'D') && (tag[2] == '3')) {  // this track has an ID3v2 tag
						// calculate tag size and read into new buffer
						int iTagSize = tag[8] * 128 + tag[9];
						tag = new char[iTagSize];
						file.Read(tag, 0, iTagSize);
						for (int i = 0; i < iTagSize - 11; i++) {
							if (tag[i] == 'T') {
								int iTagSubsectionSize = tag[i+6] * 128 + tag[i+7] - 1;
								if ((tag[i+1] == 'P') && (tag[i+2] == 'E') && (tag[i+3] == '1')) {  // found an artist name tag
									if (iTagSize < i + iTagSubsectionSize) {
										astrNewProgressMessages.Add("Error processing file " + strFilename + ": Artist tag exceeded bounds.");
									} else {
										track.strArtist = "";
										for (int j = 0; j < iTagSubsectionSize; j++)
											track.strArtist += tag[i + j + 11];
										i += iTagSubsectionSize + 10;
									}
								}
								else if ((tag[i+1] == 'I') && (tag[i+2] == 'T') && (tag[i+3] == '2')) {  // found a track name tag
									if (iTagSize < i + iTagSubsectionSize) {
										astrNewProgressMessages.Add("Error processing file " + strFilename + ": Title tag exceeded bounds.");
									} else {
										track.strTrack = "";
										for (int j = 0; j < iTagSubsectionSize; j++)
											track.strTrack += tag[i + j + 11];
										i += iTagSubsectionSize + 10;
									}
								}
								else if ((tag[i+1] == 'A') && (tag[i+2] == 'L') && (tag[i+3] == 'B')) {  // found an album name tag
									if (iTagSize < i + iTagSubsectionSize) {
										astrNewProgressMessages.Add("Error processing file " + strFilename + ": Album tag exceeded bounds.");
									} else {
										track.strAlbum = "";
										for (int j = 0; j < iTagSubsectionSize; j++)
											track.strAlbum += tag[i + j + 11];
										if (track.strAlbum == "<Unknown>")
											track.strAlbum = "";
										i += iTagSubsectionSize + 10;
									}
								}
								else if ((tag[i+1] == 'R') && (tag[i+2] == 'C') && (tag[i+3] == 'K')) {  // found a track number tag
									if (iTagSize < i + iTagSubsectionSize) {
										astrNewProgressMessages.Add("Error processing file " + strFilename + ": Track info tag exceeded bounds.");
									} else {
										track.iTrackNumber = 0;
										int iTotalTrackIndex = 0;
										if ((tag[i+11] >= '0') && (tag[i+11] <= '9')) {
											track.iTrackNumber += tag[i+11] - '0';
											if ((iTagSubsectionSize >= 14) && (tag[i+12] == '/'))
												iTotalTrackIndex = 13;
											else if ((tag[i+12] >= '0') && (tag[i+12] <= '9')) {
												track.iTrackNumber = track.iTrackNumber * 10 + tag[i+12] - '0';
												if ((iTagSubsectionSize >= 15) && (tag[i+13] == '/'))
													iTotalTrackIndex = 14;
											}
											if (iTotalTrackIndex > 0) {
												track.iTrackTotal = 0;
												if ((tag[i+iTotalTrackIndex] >= '0') && (tag[i+iTotalTrackIndex] <= '9')) {
													track.iTrackTotal += tag[i+iTotalTrackIndex] - '0';
													if ((iTagSubsectionSize >= iTotalTrackIndex + 2) && (tag[i+iTotalTrackIndex+1] >= '0') && (tag[i+iTotalTrackIndex+1] <= '9')) {
														track.iTrackTotal = track.iTrackTotal * 10 + tag[i+iTotalTrackIndex+1] - '0';
													}
												}
											}
										}
									}
									i += iTagSubsectionSize + 10;
								}
							}
						}
						if (track.strArtist == "")
							track.strFilename = "Anonymous - " + track.strTrack + ".mp3";
						else
							track.strFilename = track.strArtist + " - " + track.strTrack + ".mp3";
					}
				} catch (Exception ex) {
					astrNewProgressMessages.Add("Error processing file " + strFilename + ": " + ex.Message);
				} finally {
					file.Close();
				}
			}
			return track;
		}

		#endregion

		#region Helper functions

		private void RecursiveFilenameParse(ArrayList aFilenameTokens, int iTokenPointer, ArrayList aParameters, int iParameterPointer, string strCurrentSolution, ref ArrayList aSolutions) {
			// This function creates a list of valid track data sets from filename tokens. More than one data set may be possible, due to the use of "optional" fields. Recursion is involved here.
			if (iTokenPointer == aFilenameTokens.Count) {
				while ((iParameterPointer < aParameters.Count) && (((string) aParameters[iParameterPointer]).EndsWith("(Optional)")))
					iParameterPointer++;
			}
			if ((iTokenPointer == aFilenameTokens.Count) && (iParameterPointer == aParameters.Count)) {
				aSolutions.Add(strCurrentSolution);
			} else if ((iTokenPointer < aFilenameTokens.Count) && (iParameterPointer == aParameters.Count) && (strCurrentSolution.EndsWith("T"))) {
				aSolutions.Add(strCurrentSolution);
			} else if ((iTokenPointer < aFilenameTokens.Count) && (iParameterPointer < aParameters.Count)) {
				string strToken = (string) aFilenameTokens[iTokenPointer];
				// find qualities of token
				bool isInteger = true;
				for (int i = 0; i < strToken.Length; i++) {
					if (Char.IsDigit(strToken[i]) == false)
						isInteger = false;
				}
				bool isUnder100 = isInteger & (strToken.Length > 0) & (strToken.Length < 3);
				string strParameter = (string) aParameters[iParameterPointer];
				if (strParameter.EndsWith("(Optional)"))
					RecursiveFilenameParse(aFilenameTokens, iTokenPointer, aParameters, iParameterPointer + 1, strCurrentSolution, ref aSolutions);  // try without this parameter
				if (strParameter.StartsWith("Track Name"))
					RecursiveFilenameParse(aFilenameTokens, iTokenPointer + 1, aParameters, iParameterPointer + 1, strCurrentSolution + "T", ref aSolutions);
				else if (strParameter.StartsWith("Album"))
					RecursiveFilenameParse(aFilenameTokens, iTokenPointer + 1, aParameters, iParameterPointer + 1, strCurrentSolution + "A", ref aSolutions);
				else if (strParameter.StartsWith("Artist"))
					RecursiveFilenameParse(aFilenameTokens, iTokenPointer + 1, aParameters, iParameterPointer + 1, strCurrentSolution + "R", ref aSolutions);
				else if (strParameter.StartsWith("Comment"))   // add now
					RecursiveFilenameParse(aFilenameTokens, iTokenPointer + 1, aParameters, iParameterPointer + 1, strCurrentSolution + "C", ref aSolutions);
				else if ((strParameter.StartsWith("Year")) && (isInteger))
					RecursiveFilenameParse(aFilenameTokens, iTokenPointer + 1, aParameters, iParameterPointer + 1, strCurrentSolution + "Y", ref aSolutions);
				else if ((strParameter.StartsWith("Track Number")) && (isUnder100))
					RecursiveFilenameParse(aFilenameTokens, iTokenPointer + 1, aParameters, iParameterPointer + 1, strCurrentSolution + "#", ref aSolutions);
				else if ((strParameter.StartsWith("Genre")) && (isUnder100))
					RecursiveFilenameParse(aFilenameTokens, iTokenPointer + 1, aParameters, iParameterPointer + 1, strCurrentSolution + "G", ref aSolutions);
				else if (strParameter == "-")  // try just skipping this token
					RecursiveFilenameParse(aFilenameTokens, iTokenPointer + 1, aParameters, iParameterPointer + 1, strCurrentSolution + "-", ref aSolutions);
			}
		}

		public ArrayList ParseSolution(ArrayList aFilenameTokens, string strSolution) {
			// This function applies a solution string to a filename to tokenize it into track data.
			ArrayList aFinalSolution = new ArrayList();
			// setup array list
			aFinalSolution.Add("");
			aFinalSolution.Add("");
			aFinalSolution.Add("");
			aFinalSolution.Add((int) 0);
			aFinalSolution.Add((int) 0);
			aFinalSolution.Add((int) 0);
			aFinalSolution.Add("");
			for (int i = 0; i < strSolution.Length; i++) {
				char c = strSolution[i];
				if (c == 'R') {
					aFinalSolution[0] = (string) aFilenameTokens[i];
				} else if (c == 'T') {  // title
					aFinalSolution[1] = (string) aFilenameTokens[i];
					if (i == strSolution.Length - 1) {
						for (i++; i < aFilenameTokens.Count; i++)
							aFinalSolution[1] += " - " + aFilenameTokens[i];
					}
				} else if (c == 'A') {
					aFinalSolution[2] = (string) aFilenameTokens[i];
			 } else if (c == '#') {
					aFinalSolution[3] = Int32.Parse((string) aFilenameTokens[i]);
				} else if (c == 'Y') {
					aFinalSolution[4] = Int32.Parse((string) aFilenameTokens[i]);
				} else if (c == 'G') {
					aFinalSolution[5] = Int32.Parse((string) aFilenameTokens[i]);
				} else if (c == 'C') {  // add all remaining items
					aFinalSolution[6] = (string) aFilenameTokens[i++];
					while (i < aFilenameTokens.Count)
						aFinalSolution[6] += " " + aFilenameTokens[i++];
				}
			}
			return aFinalSolution;
		}

		private Album FindAlbum(ArrayList aAlbums, string strFilename) {
			foreach (Album album in aAlbums) {
				foreach (string strTrack in album.aTracks) {
					if (strFilename == strTrack)
						return album;
				}
			}
			return null;
		}

		private int FindTrackNumber(Album album, string strFilename) {
			for (int i = 0; i < album.aTracks.Count; i++) {
				string strTrack = (string) album.aTracks[i];
				if (strFilename == strTrack)
					return (i + 1);
			}
			return 0;
		}

		public void WriteID3v2TrackPart(BinaryWriter file, string strTag, string strValue) {
			// write artist
			if ((strValue.Length > 0) && (strTag.Length == 4)) {
				WriteStringToFile(file, strTag);
				long lTagLength = strValue.Length + 1;  // +1 is because ID3v2 requires a zero between two-byte "flags" field and tag data
				for (int power = 3; power >= 0; power--)  // write tag length (four bytes, most significant byte first)
					file.Write((char) ((lTagLength / Math.Pow(128, power)) % 128));  // note: only top 7 bits in each byte
				// write flags field and intervening zero
				for (int i = 0; i < 3; i++)
					file.Write((char) 0);
				WriteStringToFile(file, strValue);
			}
		}

		private void ActivateButtons(bool activate) {
			btnProcessFileFolder.Enabled = btnProcessSelectedFiles.Enabled = btnProcessFile.Enabled = btnProcessFolder.Enabled = activate;
			btnProcessFileFolder.Refresh();
			btnProcessSelectedFiles.Refresh();
			btnProcessFile.Refresh();
			btnProcessFolder.Refresh();
			btnAbort.Enabled = !activate;
			btnAbort.Refresh();
		}

		public void WriteStringToFile(BinaryWriter file, string strText) {
			for (int i = 0; i < strText.Length; i++)
				file.Write((char) strText[i]);
		}

		#endregion

		#region Button-click event handlers

		private void btnAdd_Click(object sender, System.EventArgs e) {
			if (lstAvailableParameters.SelectedIndices.Count == 1)
				lstParameters.Items.Add(lstAvailableParameters.SelectedItem.ToString() + (checkOptional.CheckState == CheckState.Checked ? " (Optional)" : ""));
		}

		private void lstAvailableParameters_SelectedIndexChanged(object sender, System.EventArgs e) {
			btnAdd.Enabled = (lstAvailableParameters.SelectedIndices.Count == 0 ? false : true);
		}

		private void lstParameters_SelectedIndexChanged(object sender, System.EventArgs e) {
			if (lstParameters.SelectedIndices.Count == 0) {
				btnMoveUp.Enabled = btnMoveDown.Enabled = btnDelete.Enabled = false;
			}
			else {
				btnDelete.Enabled = true;
				btnMoveUp.Enabled = (lstParameters.SelectedIndex == 0 ? false : true);
				btnMoveDown.Enabled = (lstParameters.SelectedIndex == lstParameters.Items.Count - 1 ? false : true);
			}
		}

		private void btnMoveUp_Click(object sender, System.EventArgs e) {
			if (lstParameters.SelectedIndex > 0) {
				lstParameters.BeginUpdate();
				string strMoveParameter = lstParameters.SelectedItem.ToString();
				int iNewPosition = lstParameters.SelectedIndex - 1;
				lstParameters.Items.RemoveAt(lstParameters.SelectedIndex);
				lstParameters.Items.Insert(iNewPosition, strMoveParameter);
				lstParameters.SelectedIndex = iNewPosition;
				lstParameters.EndUpdate();
			}
		}

		private void btnMoveDown_Click(object sender, System.EventArgs e) {
			if (lstParameters.SelectedIndex < lstParameters.Items.Count - 1) {
				lstParameters.BeginUpdate();
				string strMoveParameter = lstParameters.SelectedItem.ToString();
				int iNewPosition = lstParameters.SelectedIndex + 1;
				lstParameters.Items.RemoveAt(lstParameters.SelectedIndex);
				lstParameters.Items.Insert(iNewPosition, strMoveParameter);
				lstParameters.SelectedIndex = iNewPosition;
				lstParameters.EndUpdate();
			}
		}

		private void btnDelete_Click(object sender, System.EventArgs e) {
			if (lstParameters.SelectedIndex > -1) {
				lstParameters.BeginUpdate();
				string strMoveParameter = lstParameters.SelectedItem.ToString();
				int iRemovePosition = lstParameters.SelectedIndex;
				lstParameters.Items.RemoveAt(lstParameters.SelectedIndex);
				if (lstParameters.Items.Count == 0) {  // disable buttons
					btnDelete.Enabled = btnMoveUp.Enabled = btnMoveDown.Enabled = false;
				}
				else if (iRemovePosition == lstParameters.Items.Count)
					lstParameters.SelectedIndex = iRemovePosition - 1;
				else
					lstParameters.SelectedIndex = iRemovePosition;
				lstParameters.EndUpdate();
			}
		}

		private void btnApply_Click(object sender, System.EventArgs e) {
			OpenFileDialog o = new OpenFileDialog();
			o.Filter = "MP3 Files (*.mp3)|*.mp3";
			if (o.ShowDialog() == DialogResult.OK)
				ProcessMP3(o.FileName, txtArtist.Text, txtTrackName.Text, txtAlbum.Text, (int) nudTrackNumber.Value, (int) nudTrackTotal.Value, (int) nudYear.Value, (int) nudGenre.Value, txtComment.Text);
		}

		private void nudTrackNumber_ValueChanged(object sender, System.EventArgs e) {
			nudTrackTotal.Value = Math.Min(nudTrackTotal.Value, nudTrackNumber.Value);
		}

		private void nudGenre_ValueChanged(object sender, System.EventArgs e) {
			if (nudGenre.Value < cmbGenre.Items.Count)
				cmbGenre.SelectedIndex = (int) nudGenre.Value;
		}

		private void cmbGenre_SelectedIndexChanged(object sender, System.EventArgs e) {
			nudGenre.Value = cmbGenre.SelectedIndex;
		}

		private void btnProcessSelectedFiles_Click(object sender, System.EventArgs e) {
			OpenFileDialog d = new OpenFileDialog();
			d.Filter = "MP3 Files (*.mp3)|*.mp3";
			d.Multiselect = true;
			try {
				if (d.ShowDialog() == DialogResult.OK) {
					strFilenames = new ArrayList();
					foreach (string s in d.FileNames)
						strFilenames.Add(s);
					processing = new Thread(new ThreadStart(ProcessTags));
					processing.Start();
				}
			} catch (Exception ex) {
				MessageBox.Show(ex.Message, "Error");
			}
		}

		private void btnProcessFileFolder_Click(object sender, System.EventArgs e) {
			FolderBrowserDialog f = new FolderBrowserDialog();
			if (f.ShowDialog() == DialogResult.OK) {
				string[] strFiles = System.IO.Directory.GetFiles(f.SelectedPath, "*.mp3");
				strFilenames = new ArrayList();
				foreach (string strFile in strFiles)
					strFilenames.Add(strFile);
				if (strFilenames.Count > 0) {
					processing = new Thread(new ThreadStart(ProcessTags));
					processing.Start();
				}
			}
		}

		private void btnBrowseFile_Click(object sender, System.EventArgs e) {
			OpenFileDialog file = new OpenFileDialog();
			file.Filter = "MP3 Files (*.mp3)|*.mp3";
			if (file.ShowDialog() == DialogResult.OK)
				txtFilename.Text = file.FileName;
		}

		private void btnBrowseFolder_Click(object sender, System.EventArgs e) {
			FolderBrowserDialog folder = new FolderBrowserDialog();
			folder.RootFolder = Environment.SpecialFolder.Desktop;
			if (folder.ShowDialog() == DialogResult.OK)
				txtFolder.Text = folder.SelectedPath;
		}

		private void txtFilename_TextChanged(object sender, System.EventArgs e) {
			btnProcessFile.Enabled = System.IO.File.Exists(txtFilename.Text);
		}

		private void txtFolder_TextChanged(object sender, System.EventArgs e) {
			btnProcessFolder.Enabled = System.IO.Directory.Exists(txtFolder.Text);
		}

		private void btnProcessFile_Click(object sender, System.EventArgs e) {
			strPath = txtFilename.Text.Substring(0, txtFilename.Text.LastIndexOf(@"\"));
			strFilenames = new ArrayList();
			strFilenames.Add(txtFilename.Text);
			processing = new Thread(new ThreadStart(ProcessFilenames));
			processing.Start();
		}

		private void btnProcessFolder_Click(object sender, System.EventArgs e) {
			strPath = txtFolder.Text;
			string[] strFiles = Directory.GetFiles(strPath, "*.mp3");
			strFilenames = new ArrayList();
			foreach (string strFile in strFiles)
				strFilenames.Add(strFile);
			processing = new Thread(new ThreadStart(ProcessFilenames));
			processing.Start();
		}

		private void btnAbort_Click(object sender, System.EventArgs e) {
			if (processing != null) {
				bContinue = false;
				astrNewProgressMessages.Add("Task aborted.");
			}
		}

		private void btnHelp_Click(object sender, System.EventArgs e) {
			HelpDialog h = new HelpDialog();
			h.ShowDialog();
		}

		#endregion

		#region Other event handlers

		private void timerUpdate_Tick(object sender, EventArgs e) {
			lock (astrNewProgressMessages) {
				if (astrNewProgressMessages.Count > 0) {
					lstProgress.BeginUpdate();
					while (astrNewProgressMessages.Count > 0) {
						lstProgress.Items.Add(astrNewProgressMessages[0]);
						astrNewProgressMessages.RemoveAt(0);
					}
					while (lstProgress.Items.Count > 100)
						lstProgress.Items.RemoveAt(0);
					lstProgress.EndUpdate();
					lstProgress.Refresh();
				}
			}
			pbProgress.Value = iProgress;
			grpProgress.Text = strStatus;
			ActivateButtons(processing == null ? true : false);
		}

		private void MP3ID3Converter_Closing(object sender, System.ComponentModel.CancelEventArgs e) {
			if (processing != null) {
				processing.Abort();
				processing = null;
			}
		}

		#endregion

	}

	#region Assistant classes (enums, data-structure classes, etc.)

	public enum Parameters { Artist, TrackName, Album, TrackNo, Year, Genre };

	public class Album {
		public string strName = "";
		public ArrayList aTracks;
		public int iTracks = 0;
		public int iTotalTracks = 0;
		public Album() {
			aTracks = new ArrayList();
		}
	}

	public class Playlist {
		public string strPlaylistName = "";
		public ArrayList aTracks;
		public Playlist() {
			aTracks = new ArrayList();
		}
	}

	public class Track {
		public string strArtist = "";
		public string strTrack = "";
		public string strAlbum = "";
		public string strFilename = "";
		public int iTrackNumber = 0;
		public int iTrackTotal = 0;
	}

	#endregion

}
